#include "usg_common.h"
#include "hashtable.h"
#include "mm.h"
#include "svsem.h"
#include "logger_factory.h"
#include <set>
#include <functional>

typedef struct tailq_entry_t
{
  void *value;
  int key;
  /*
   * This holds the pointers to the next and previous joint in
   * the tail queue.
   */
  TAILQ_ENTRY(tailq_entry_t) joint;
} tailq_entry_t;

#define START_KEY 1000

typedef  TAILQ_HEAD(tailq_header_t, tailq_entry_t) tailq_header_t;


static size_t hashcode(int key);

static struct timespec TIMEOUT = {2, 0};

void hashtable_init(hashtable_t *hashtable )
{

  memset(hashtable, 0, sizeof(hashtable_t));
  hashtable->mutex = svsem_get(IPC_PRIVATE, 1);
  hashtable->wlock = svsem_get(IPC_PRIVATE, 1);
  hashtable->cond = svsem_get(IPC_PRIVATE, 1);
  hashtable->readcnt = 0;

  FILE * semfile = fopen("./sem.txt", "w+");
  if(semfile == NULL) {
    err_exit(errno, "fopen");
  }
  fprintf(semfile, "hashtable->mutex=%d\n", hashtable->mutex);
  fclose(semfile);
}

void hashtable_destroy(hashtable_t *hashtable) {
  svsem_remove( hashtable->mutex);
  svsem_remove( hashtable->wlock);
  svsem_remove( hashtable->cond);
}


static inline void *_hashtable_get(hashtable_t *hashtable, int key)
{
  size_t code = hashcode(key);
  tailq_entry_t *item;
  tailq_header_t *my_tailq_head = hashtable->array[code] ;
  if ( my_tailq_head == NULL)
  {
    return NULL;
  }
  else
  {

    TAILQ_FOREACH(item, my_tailq_head, joint)
    {
      if (key == item->key)
        return item->value;
    }
  }
  return NULL;

}

static inline void * _hashtable_put(hashtable_t *hashtable, int key, void *value)
{
  size_t code = hashcode(key);
  void *oldvalue;
  tailq_entry_t *item;
  tailq_header_t *my_tailq_head = hashtable->array[code] ;
  if ( my_tailq_head == NULL)
  {
    my_tailq_head  = (tailq_header_t*) mm_malloc(sizeof(tailq_header_t ));
    TAILQ_INIT(my_tailq_head);
    hashtable->array[code] = my_tailq_head;
    goto putnew;
  }

  TAILQ_FOREACH(item, my_tailq_head, joint)
  {
    if (key ==item->key)
    {
      oldvalue = item -> value;
      item->key= key;
      item -> value = value;
      return oldvalue;
    }
  }
putnew:
  item = (tailq_entry_t *) mm_malloc(sizeof(tailq_entry_t));
  item->key = key;
  item -> value = value;
  TAILQ_INSERT_TAIL(my_tailq_head, item, joint);
  return NULL;
}

void *hashtable_remove(hashtable_t *hashtable, int key)
{
  size_t code = hashcode(key);
  tailq_entry_t *item;
  void *oldvalue;
  int rv;

  if( (rv = svsem_wait(hashtable->wlock)) != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_remove\n");
  }
  tailq_header_t *my_tailq_head = hashtable->array[code] ;
  if ( my_tailq_head == NULL)
  {
    svsem_post(hashtable->wlock);
    return NULL;
  }
  else
  {
    for (item = TAILQ_FIRST(my_tailq_head); item != NULL; item = TAILQ_NEXT(item, joint))
    {
      if (key == item->key)
      {
        oldvalue = item->value;
        /* Remove the item from the tail queue. */
        TAILQ_REMOVE(my_tailq_head, item, joint);

        /* mm_free the item as we don't need it anymore. */
        mm_free(item);
        svsem_post(hashtable->wlock);
        return oldvalue;
      }
    }
  }
  if((rv = svsem_post(hashtable->wlock)) != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_remove\n");
  }

  return NULL;

}





void hashtable_removeall(hashtable_t *hashtable)
{
  tailq_entry_t *item;
  int rv;
  rv = svsem_wait(hashtable->wlock);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_removeall\n");
  }
  for (int i = 0; i < MAPSIZE; i++)
  {
    tailq_header_t *my_tailq_head = hashtable->array[i] ;

    if (my_tailq_head == NULL )
      continue;

    while ((item = TAILQ_FIRST(my_tailq_head)) )
    {
      TAILQ_REMOVE(my_tailq_head, item, joint);
      mm_free(item);
    }
    mm_free(my_tailq_head);
    hashtable->array[i] = NULL;
  }
  rv = svsem_post(hashtable->wlock);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_removeall\n");
  }
}

/**
 * for debug
 */
void hashtable_printall(hashtable_t *hashtable)
{
  tailq_entry_t *item;
  for (int i = 0; i < MAPSIZE; i++)
  {
    tailq_header_t *my_tailq_head = hashtable->array[i] ;

    if (my_tailq_head == NULL )
      continue;

    printf("code=%d\n", i);
    TAILQ_FOREACH(item, my_tailq_head, joint)
    {
      printf("%d:%s\n", item->key, (char *)item->value);
    }
    printf("\n");
  }
}

static size_t hashcode(int key)
{

  return key % MAPSIZE;
  /*printf("hashfun = %ld\n", code);*/
}

void *hashtable_get(hashtable_t *hashtable, int key) {
LoggerFactory::getLogger()->debug( "==========hashtable_get before 1");
  
  int rv;
  rv = svsem_wait(hashtable->mutex);

LoggerFactory::getLogger()->debug( "==========hashtable_get before 2");

  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_get 1");
  }
  hashtable->readcnt++;
  if (hashtable->readcnt == 1) {
    //获取读写锁
LoggerFactory::getLogger()->debug( "==========hashtable_get before 3");
    rv = svsem_wait(hashtable->wlock);
LoggerFactory::getLogger()->debug( "==========hashtable_get before 4");
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_get 2");
    }
  }
  rv = svsem_post(hashtable->mutex);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_get 3");
  }
  // ================

  void * res = _hashtable_get(hashtable, key);

  // ==================

  rv = svsem_wait(hashtable->mutex);
LoggerFactory::getLogger()->debug( "==========hashtable_get before 5");
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_get 4");
  }
  hashtable->readcnt--;
  if(hashtable->readcnt == 0) {
    //释放读写锁

    rv = svsem_post(hashtable->wlock);
LoggerFactory::getLogger()->debug( "==========hashtable_get before 6");
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_get 5");
    }
    //通知写
    rv = svsem_set(hashtable->cond, 1);
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_get 6");
    }
  }

  rv = svsem_post(hashtable->mutex);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_get 7");
  }
LoggerFactory::getLogger()->debug( "==========hashtable_get after");
  return res;
}

void hashtable_put(hashtable_t *hashtable, int key, void *value) {

  int rv;
LoggerFactory::getLogger()->debug( "==========hashtable_put before 1");
  rv = svsem_wait(hashtable->mutex);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_put\n");
  }
LoggerFactory::getLogger()->debug( "==========hashtable_put before 2");
  // 设置读优先级高
  while (hashtable->readcnt > 0)
  {
    rv = svsem_set(hashtable->cond, 0);
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_put\n");
    }
    rv = svsem_post(hashtable->mutex);
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_put\n");
    }
    //等待写通知
LoggerFactory::getLogger()->debug( "==========hashtable_put before 3");
    rv = svsem_wait(hashtable->cond);
LoggerFactory::getLogger()->debug( "==========hashtable_put before 4");
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_put\n");
    }


    rv = svsem_wait(hashtable->mutex);
LoggerFactory::getLogger()->debug( "==========hashtable_put before 5");
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_put\n");
    }
  }
  rv = svsem_post(hashtable->mutex);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_put\n");
  }

  //获取读写锁
  rv = svsem_wait(hashtable->wlock);
LoggerFactory::getLogger()->debug( "==========hashtable_put before 6");
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_put\n");
  }

  _hashtable_put(hashtable, key, value);

  //释放读写锁
  rv = svsem_post(hashtable->wlock);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_put\n");
  }

  LoggerFactory::getLogger()->debug( "==========hashtable_put after");
}



static inline void _hashtable_foreach(hashtable_t *hashtable, std::function<void(int, void *)>  cb) {
  tailq_entry_t *item;
  for (int i = 0; i < MAPSIZE; i++) {
    tailq_header_t *my_tailq_head = hashtable->array[i] ;

    if (my_tailq_head == NULL )
      continue;

    TAILQ_FOREACH(item, my_tailq_head, joint)
    {
      cb(item->key,  item -> value);
    }
  }
}


void hashtable_foreach(hashtable_t *hashtable,  std::function<void(int, void *)>  cb) {
  int rv;


  LoggerFactory::getLogger()->debug("===hashtable_foreach before 1\n");


  rv = svsem_wait(hashtable->mutex);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_foreach");
  }
  hashtable->readcnt++;
  if (hashtable->readcnt == 1) {
    //获取读写锁
    rv = svsem_wait(hashtable->wlock);
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_foreach");
    }
  }
  rv = svsem_post(hashtable->mutex);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_foreach");
  }

  // ==================

  _hashtable_foreach(hashtable, cb);

  // ==================

  rv = svsem_wait(hashtable->mutex);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_foreach");
  }
  hashtable->readcnt--;
  if(hashtable->readcnt == 0) {
    //释放读写锁
    rv = svsem_post(hashtable->wlock);
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_foreach");
    }
    //通知写
    rv = svsem_set(hashtable->cond, 1);
    if(rv != 0) {
      LoggerFactory::getLogger()->error(rv, "hashtable_foreach");
    }
  }
  rv = svsem_post(hashtable->mutex);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_foreach");
  }

}


std::set<int> * hashtable_keyset(hashtable_t *hashtable) {
  std::set<int> *keyset = new std::set<int>;
  tailq_entry_t *item;
  for (int i = 0; i < MAPSIZE; i++) {
    tailq_header_t *my_tailq_head = hashtable->array[i] ;

    if (my_tailq_head == NULL )
      continue;

    TAILQ_FOREACH(item, my_tailq_head, joint)
    {
      keyset->insert(item->key);
    }
  }
  return keyset;
}


int hashtable_alloc_key(hashtable_t *hashtable) {
  int rv;
  int key = START_KEY;
  rv = svsem_wait(hashtable->wlock);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_alloc_key\n");
  }

  while(_hashtable_get(hashtable, key) != NULL) {
    key++;
  }
  // 占用key
  _hashtable_put(hashtable, key, (void *)1);

  rv = svsem_post(hashtable->wlock);
  if(rv != 0) {
    LoggerFactory::getLogger()->error(rv, "hashtable_alloc_key\n");
  }
  return key;
}
